from typing import Any, Callable

import httpx
import llsd
import pytest
from starlette.requests import Request
from starlette.responses import JSONResponse, PlainTextResponse, Response
from starlette.types import Receive, Scope, Send

from llsd_asgi import LLSDMiddleware
from tests.utils import mock_receive, mock_send

Format = Callable[[Any], bytes]
Parse = Callable[[bytes], Any]


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "content_type,format",
    [
        ("application/llsd+xml", llsd.format_xml),
        ("application/llsd+binary", llsd.format_binary),
        ("application/llsd+notation", llsd.format_notation),
    ],
)
async def test_llsd_request(content_type: str, format: Format) -> None:
    async def app(scope: Scope, receive: Receive, send: Send) -> None:
        request = Request(scope, receive=receive)
        content_type = request.headers["content-type"]
        data = await request.json()
        message = data["message"]
        text = f"content_type={content_type!r} message={message!r}"

        response = PlainTextResponse(text)
        await response(scope, receive, send)

    app = LLSDMiddleware(app)

    async with httpx.AsyncClient(app=app, base_url="http://testserver") as client:
        content = {"message": "Hello, world!"}
        body = format(content)
        r = await client.post("/", content=body, headers={"content-type": content_type})
        assert r.status_code == 200
        assert r.text == "content_type='application/json' message='Hello, world!'"


@pytest.mark.asyncio
async def test_non_llsd_request() -> None:
    async def app(scope: Scope, receive: Receive, send: Send) -> None:
        request = Request(scope, receive=receive)
        content_type = request.headers["content-type"]
        message = (await request.body()).decode()
        text = f"content_type={content_type!r} message={message!r}"

        response = PlainTextResponse(text)
        await response(scope, receive, send)

    app = LLSDMiddleware(app)

    async with httpx.AsyncClient(app=app, base_url="http://testserver") as client:
        r = await client.post(
            "/",
            content="Hello, world!",
            headers={"content-type": "text/plain"},
        )
        assert r.status_code == 200
        assert r.text == "content_type='text/plain' message='Hello, world!'"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "header,format,parse",
    [
        ("application/llsd+xml", llsd.format_xml, llsd.parse_xml),
        ("application/llsd+binary", llsd.format_binary, llsd.parse_binary),
        ("application/llsd+notation", llsd.format_notation, llsd.parse_notation),
    ],
)
async def test_llsd_accepted(header: str, format: Format, parse: Parse) -> None:
    app = LLSDMiddleware(JSONResponse({"message": "Hello, world!"}))

    async with httpx.AsyncClient(app=app, base_url="http://testserver") as client:
        r = await client.get("/", headers={"accept": header})
        assert r.status_code == 200
        assert r.headers["content-type"] == header
        expected_data = {"message": "Hello, world!"}
        assert int(r.headers["content-length"]) == len(format(expected_data))
        assert parse(r.content) == expected_data


@pytest.mark.asyncio
async def test_llsd_accepted_but_response_is_not_json() -> None:
    app = LLSDMiddleware(PlainTextResponse("Hello, world!"))

    async with httpx.AsyncClient(app=app, base_url="http://testserver") as client:
        r = await client.get("/", headers={"accept": "application/llsd+xml"})
        assert r.status_code == 200
        assert r.headers["content-type"] == "text/plain; charset=utf-8"
        assert r.text == "Hello, world!"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "header,format,parse",
    [
        ("application/llsd+xml", llsd.format_xml, llsd.parse_xml),
        ("application/llsd+binary", llsd.format_binary, llsd.parse_binary),
        ("application/llsd+notation", llsd.format_notation, llsd.parse_notation),
    ],
)
async def test_llsd_accepted_and_response_is_already_llsd(header: str, format: Format, parse: Parse) -> None:
    data = format({"message": "Hello, world!"})
    response = Response(data, media_type=header)
    app = LLSDMiddleware(response)

    async with httpx.AsyncClient(app=app, base_url="http://testserver") as client:
        r = await client.get("/", headers={"accept": header})
        assert r.status_code == 200
        assert r.headers["content-type"] == header
        expected_data = {"message": "Hello, world!"}
        assert int(r.headers["content-length"]) == len(format(expected_data))
        assert parse(r.content) == expected_data


@pytest.mark.asyncio
async def test_llsd_not_accepted() -> None:
    app = LLSDMiddleware(JSONResponse({"message": "Hello, world!"}))

    async with httpx.AsyncClient(app=app, base_url="http://testserver") as client:
        r = await client.get("/")
        assert r.status_code == 200
        assert r.headers["content-type"] == "application/json"
        assert r.json() == {"message": "Hello, world!"}
        with pytest.raises(llsd.LLSDParseError):
            llsd.parse_xml(r.content)


@pytest.mark.asyncio
async def test_request_is_not_http() -> None:
    async def lifespan_only_app(scope: Scope, receive: Receive, send: Send) -> None:
        assert scope["type"] == "lifespan"

    app = LLSDMiddleware(lifespan_only_app)
    scope = {"type": "lifespan"}
    await app(scope, mock_receive, mock_send)


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "accept",
    [(None), ("*/*"), ("text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,*/*;q=0.8")],
)
async def test_quirks_encode(accept: str) -> None:
    app = LLSDMiddleware(JSONResponse({"message": "Hello, world!"}), quirks=True)

    async with httpx.AsyncClient(app=app, base_url="http://testserver") as client:
        headers = {}
        if accept is None:
            # Emulate a client that doesn't send an Accept header
            del client._headers["accept"]
        else:
            headers["accept"] = accept
        r = await client.get("/")
        assert r.status_code == 200
        assert "content-type" not in r.headers
        assert llsd.parse_xml(r.content) == {"message": "Hello, world!"}


@pytest.mark.asyncio
async def test_quirks_decode():
    async def app(scope: Scope, receive: Receive, send: Send) -> None:
        request = Request(scope, receive=receive)
        data = await request.json()
        message = data["message"]
        text = f"message={message!r}"

        response = PlainTextResponse(text)
        await response(scope, receive, send)

    app = LLSDMiddleware(app, quirks=True)

    async with httpx.AsyncClient(app=app, base_url="http://testserver") as client:
        r = await client.post("/", content=llsd.format_xml({"message": "Hello, world!"}))
        assert r.status_code == 200
        assert r.text == "message='Hello, world!'"


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "accept",
    [("application/json,*/*")],
)
async def test_quirks_exceptions(accept: str) -> None:
    app = LLSDMiddleware(JSONResponse({"message": "Hello, world!"}), quirks=True)

    async with httpx.AsyncClient(app=app, base_url="http://testserver") as client:
        client.headers["accept"] = accept
        r = await client.get("/")
        assert r.status_code == 200
        assert r.headers["content-type"] == "application/json"
        assert r.json() == {"message": "Hello, world!"}



import json

import llsd
from starlette.datastructures import Headers, MutableHeaders
from starlette.types import ASGIApp, Message, Receive, Scope, Send


class LLSDMiddleware:
    def __init__(self, app: ASGIApp, quirks: bool = False) -> None:
        self.app = app
        self.quirks = quirks

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] == "http":
            responder = _LLSDResponder(self.app, self.quirks)
            await responder(scope, receive, send)
            return
        await self.app(scope, receive, send)


_CONTENT_TYPE_TO_PARSE = {
    "application/llsd+xml": llsd.parse_xml,
    "application/llsd+binary": llsd.parse_binary,
    "application/llsd+notation": llsd.parse_notation,
}

_CONTENT_TYPE_TO_FORMAT = {
    "application/llsd+xml": llsd.format_xml,
    "application/llsd+binary": llsd.format_binary,
    "application/llsd+notation": llsd.format_notation,
}


class _LLSDResponder:

    def __init__(self, app: ASGIApp, quirks: bool = False) -> None:
        self.app = app
        self.quirks = quirks
        self.should_decode_from_llsd_to_json = False
        self.should_encode_from_json_to_llsd = False
        self.receive: Receive = unattached_receive
        self.send: Send = unattached_send
        self.initial_message: Message = {}
        self.started = False

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        headers = MutableHeaders(scope=scope)

        try:
            self.parse = _CONTENT_TYPE_TO_PARSE[
                headers.get("content-type", "application/llsd+xml" if self.quirks else None)
            ]
            self.should_decode_from_llsd_to_json = True
        except KeyError:
            self.should_decode_from_llsd_to_json = False

        try:
            self.accept_header = headers.get("accept")
            self.format = _CONTENT_TYPE_TO_FORMAT[self.accept_header]
            self.should_encode_from_json_to_llsd = True
        except KeyError:
            # Quirks mode matches the behavior of some poorly behaved clients,
            # which expect llsd responses even if they don't send an Accept header.
            if self.quirks:
                if not self.accept_header or (
                    "*/*" in self.accept_header and "application/json" not in self.accept_header
                ):
                    self.format = llsd.format_xml
                    self.should_encode_from_json_to_llsd = True
                    self.accept_header = None
            else:
                self.should_encode_from_json_to_llsd = False

        self.receive = receive
        self.send = send

        if self.should_decode_from_llsd_to_json:
            # We're going to present JSON content to the application,
            # so rewrite `Content-Type` for consistency and compliance
            # with possible downstream security checks in some frameworks.
            # See: https://github.com/florimondmanca/msgpack-asgi/issues/23
            headers["content-type"] = "application/json"

        await self.app(scope, self.receive_with_llsd, self.send_with_llsd)

    async def receive_with_llsd(self) -> Message:
        message = await self.receive()

        if not self.should_decode_from_llsd_to_json:
            return message

        assert message["type"] == "http.request"

        body = message["body"]
        more_body = message.get("more_body", False)
        if more_body:
            # Some implementations (e.g. HTTPX) may send one more empty-body message.
            # Make sure they don't send one that contains a body, or it means
            # that clients attempt to stream the request body.
            message = await self.receive()
            if message["body"] != b"":  # pragma: no cover
                raise NotImplementedError("Streaming the request body isn't supported yet")

        message["body"] = json.dumps(self.parse(body)).encode()

        return message

    async def send_with_llsd(self, message: Message) -> None:
        if not self.should_encode_from_json_to_llsd:
            await self.send(message)
            return

        if message["type"] == "http.response.start":
            headers = Headers(raw=message["headers"])
            if headers.get("content-type") != "application/json":
                # Client accepts llsd, but the app did not send JSON data.
                # (Note that it may have sent llsd-encoded data.)
                self.should_encode_from_json_to_llsd = False
                await self.send(message)
                return

            # Don't send the initial message until we've determined how to
            # modify the ougoging headers correctly.
            self.initial_message = message

        elif message["type"] == "http.response.body":
            assert self.should_encode_from_json_to_llsd

            body = message.get("body", b"")
            more_body = message.get("more_body", False)
            if more_body:  # pragma: no cover
                raise NotImplementedError("Streaming the response body isn't supported yet")

            body = self.format(json.loads(body))

            headers = MutableHeaders(raw=self.initial_message["headers"])
            if self.accept_header:
                headers["Content-Type"] = self.accept_header
            else:
                # quirks mode allows a response without a Content-Type response header
                del headers["Content-Type"]
            headers["Content-Length"] = str(len(body))
            message["body"] = body

            await self.send(self.initial_message)
            await self.send(message)


async def unattached_receive() -> Message:
    raise RuntimeError("receive awaitable not set")  # pragma: no cover


async def unattached_send(message: Message) -> None:
    raise RuntimeError("send awaitable not set")  # pragma: no cover
